using System.Collections.Generic;
using System.Linq;
using UnityEngine;
using UnityEngine.Events;

namespace TittyMagic.UI
{
    internal class PhysicsWindow : WindowBase
    {
        private readonly JSONStorableString _jointPhysicsParamsHeader;
        private readonly JSONStorableString _softPhysicsParamsHeader;

        private UnityAction onReturnToParent => () =>
        {
            activeNestedWindow = null;
            buildAction();
        };

        public PhysicsWindow(Script script) : base(script)
        {
            buildAction = BuildSelf;

            _jointPhysicsParamsHeader = new JSONStorableString("mainPhysicsParamsHeader", "");
            if(Gender.isFemale)
            {
                _softPhysicsParamsHeader = new JSONStorableString("softPhysicsParamsHeader", "");
            }

            nestedWindows.Add(new OverviewWindow(script, onReturnToParent));
            CreateParameterWindows();
        }

        private void CreateParameterWindows()
        {

            foreach(var kvp in script.mainPhysicsHandler.parameterGroups)
            {
                nestedWindows.Add(
                    new ParameterWindow(
                        script,
                        kvp.Key,
                        script.mainPhysicsHandler.parameterGroups[kvp.Key],
                        onReturnToParent
                    )
                );
            }

            if(Gender.isFemale)
            {
                foreach(var kvp in script.softPhysicsHandler.parameterGroups)
                {
                    nestedWindows.Add(
                        new ParameterWindow(
                            script,
                            kvp.Key,
                            script.softPhysicsHandler.parameterGroups[kvp.Key],
                            onReturnToParent
                        )
                    );
                }
            }
        }

        private void BuildSelf()
        {
            CreateHeader(_jointPhysicsParamsHeader, "Joint Physics Parameters", false);
            // CreateOverviewButton(false);
            script.mainPhysicsHandler?.parameterGroups.ToList()
                .ForEach(kvp => CreateParamButton(kvp.Key, kvp.Value, false));

            if(Gender.isFemale)
            {
                CreateHeader(_softPhysicsParamsHeader, "Soft Physics Parameters", true);
                CreateAllowSelfCollisionToggle(true);
                script.softPhysicsHandler.parameterGroups.ToList()
                    .ForEach(kvp => CreateParamButton(kvp.Key, kvp.Value, true));
            }

            // string[] mainPhysicsDebug =
            //     script.mainPhysicsHandler.parameterGroups.Values.Select(v => $"{v.displayName} {v.left.valueJsf.val}").ToArray();
            // string[] softPhysicsDebug =
            //     script.softPhysicsHandler.parameterGroups.Values.Select(v => $"{v.displayName} {v.left.valueJsf.val}").ToArray();
            // Debug.Log($"Main:\n{string.Join("\n", mainPhysicsDebug)}");
            // Debug.Log($"Soft:\n{string.Join("\n", softPhysicsDebug)}");
        }

        private void CreateAllowSelfCollisionToggle(bool rightSide)
        {
            var storable = script.softPhysicsHandler.allowSelfCollision;
            var toggle = script.CreateToggle(storable, rightSide);
            toggle.height = 52;
            toggle.label = "Breast Soft Physics Self Collide";
            elements[storable.name] = toggle;
        }

        private void CreateHeader(JSONStorableString storable, string text, bool rightSide, int spacing = 0)
        {
            elements[$"{storable.name}Spacer"] = script.NewSpacer(spacing, rightSide);
            elements[storable.name] = UIHelpers.HeaderTextField(script, storable, text, rightSide);
        }

        private void CreateParamButton(string key, PhysicsParameterGroup param, bool rightSide)
        {
            var button = script.CreateButton("  " + param.displayName, rightSide);
            button.height = 52;
            button.buttonText.alignment = TextAnchor.MiddleLeft;

            var nestedWindow = (ParameterWindow) nestedWindows.Find(window => (window as ParameterWindow)?.id == key);
            button.AddListener(() =>
            {
                ClearSelf();
                activeNestedWindow = nestedWindow;
                activeNestedWindow.Rebuild();
            });

            nestedWindow.parentButton = button;
            button.label = nestedWindow.ParamButtonLabel();
            elements[key] = button;
        }

        private void CreateOverviewButton(bool rightSide)
        {
            var button = script.CreateButton("  Overview", rightSide);
            button.height = 52;
            button.buttonText.alignment = TextAnchor.MiddleLeft;

            var nestedWindow = (OverviewWindow) nestedWindows.Find(window => (window as OverviewWindow)?.id == "Overview");
            button.AddListener(() =>
            {
                ClearSelf();
                activeNestedWindow = nestedWindow;
                activeNestedWindow.Rebuild();
            });

            elements["Overview"] = button;
        }
    }
}
